@@ -9,52 +9,88 @@ from contextlib import suppress
 from http.cookies import SimpleCookie
 from itertools import cycle, islice
 from time import monotonic
+from types import TracebackType
+from typing import (TYPE_CHECKING, Any, Awaitable, Callable,  # noqa
+                    DefaultDict, Dict, Iterator, List, Optional, Set, Tuple,
+                    Type, Union, cast)
 
 import attr
 
 from . import hdrs, helpers
+from .abc import AbstractResolver
 from .client_exceptions import (ClientConnectionError,
                                 ClientConnectorCertificateError,
                                 ClientConnectorError, ClientConnectorSSLError,
                                 ClientHttpProxyError,
                                 ClientProxyConnectionError,
-                                ServerFingerprintMismatch, certificate_errors,
+                                ServerFingerprintMismatch, cert_errors,
                                 ssl_errors)
 from .client_proto import ResponseHandler
 from .client_reqrep import ClientRequest, Fingerprint, _merge_ssl_params
-from .helpers import PY_36, CeilTimeout, is_ip_address, noop, sentinel
+from .helpers import (PY_36, CeilTimeout, get_running_loop, is_ip_address,
+                      noop2, sentinel)
+from .http import RESPONSES
 from .locks import EventResultOrError
 from .resolver import DefaultResolver
 
 
 try:
     import ssl
+    SSLContext = ssl.SSLContext
 except ImportError:  # pragma: no cover
-    ssl = None
+    ssl = None  # type: ignore
+    SSLContext = object  # type: ignore
 
 
 __all__ = ('BaseConnector', 'TCPConnector', 'UnixConnector')
 
 
+if TYPE_CHECKING:  # pragma: no cover
+    from .client import ClientTimeout  # noqa
+    from .client_reqrep import ConnectionKey  # noqa
+    from .tracing import Trace  # noqa
+
+
+class _DeprecationWaiter:
+    __slots__ = ('_awaitable', '_awaited')
+
+    def __init__(self, awaitable: Awaitable[Any]) -> None:
+        self._awaitable = awaitable
+        self._awaited = False
+
+    def __await__(self) -> Any:
+        self._awaited = True
+        return self._awaitable.__await__()
+
+    def __del__(self) -> None:
+        if not self._awaited:
+            warnings.warn("Connector.close() is a coroutine, "
+                          "please use await connector.close()",
+                          DeprecationWarning)
+
+
 class Connection:
 
     _source_traceback = None
     _transport = None
 
-    def __init__(self, connector, key, protocol, loop):
+    def __init__(self, connector: 'BaseConnector',
+                 key: 'ConnectionKey',
+                 protocol: ResponseHandler,
+                 loop: asyncio.AbstractEventLoop) -> None:
         self._key = key
         self._connector = connector
         self._loop = loop
-        self._protocol = protocol
-        self._callbacks = []
+        self._protocol = protocol  # type: Optional[ResponseHandler]
+        self._callbacks = []  # type: List[Callable[[], None]]
 
         if loop.get_debug():
             self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return 'Connection<{}>'.format(self._key)
 
-    def __del__(self, _warnings=warnings):
+    def __del__(self, _warnings: Any=warnings) -> None:
         if self._protocol is not None:
             if PY_36:
                 kwargs = {'source': self}
@@ -76,33 +112,34 @@ class Connection:
             self._loop.call_exception_handler(context)
 
     @property
-    def loop(self):
+    def loop(self) -> asyncio.AbstractEventLoop:
+        warnings.warn("connector.loop property is deprecated",
+                      DeprecationWarning,
+                      stacklevel=2)
         return self._loop
 
     @property
-    def transport(self):
+    def transport(self) -> Optional[asyncio.Transport]:
+        if self._protocol is None:
+            return None
         return self._protocol.transport
 
     @property
-    def protocol(self):
+    def protocol(self) -> Optional[ResponseHandler]:
         return self._protocol
 
-    @property
-    def writer(self):
-        return self._protocol.writer
-
-    def add_callback(self, callback):
+    def add_callback(self, callback: Callable[[], None]) -> None:
         if callback is not None:
             self._callbacks.append(callback)
 
-    def _notify_release(self):
+    def _notify_release(self) -> None:
         callbacks, self._callbacks = self._callbacks[:], []
 
         for cb in callbacks:
             with suppress(Exception):
                 cb()
 
-    def close(self):
+    def close(self) -> None:
         self._notify_release()
 
         if self._protocol is not None:
@@ -110,7 +147,7 @@ class Connection:
                 self._key, self._protocol, should_close=True)
             self._protocol = None
 
-    def release(self):
+    def release(self) -> None:
         self._notify_release()
 
         if self._protocol is not None:
@@ -119,22 +156,15 @@ class Connection:
                 should_close=self._protocol.should_close)
             self._protocol = None
 
-    def detach(self):
-        self._notify_release()
-
-        if self._protocol is not None:
-            self._connector._release_acquired(self._protocol)
-        self._protocol = None
-
     @property
-    def closed(self):
+    def closed(self) -> bool:
         return self._protocol is None or not self._protocol.is_connected()
 
 
 class _TransportPlaceholder:
     """ placeholder for BaseConnector.connect function """
 
-    def close(self):
+    def close(self) -> None:
         pass
 
 
@@ -157,9 +187,12 @@ class BaseConnector:
     # abort transport after 2 seconds (cleanup broken connections)
     _cleanup_closed_period = 2.0
 
-    def __init__(self, *, keepalive_timeout=sentinel,
-                 force_close=False, limit=100, limit_per_host=0,
-                 enable_cleanup_closed=False, loop=None):
+    def __init__(self, *,
+                 keepalive_timeout: Union[object, None, float]=sentinel,
+                 force_close: bool=False,
+                 limit: int=100, limit_per_host: int=0,
+                 enable_cleanup_closed: bool=False,
+                 loop: Optional[asyncio.AbstractEventLoop]=None) -> None:
 
         if force_close:
             if keepalive_timeout is not None and \
@@ -170,23 +203,22 @@ class BaseConnector:
             if keepalive_timeout is sentinel:
                 keepalive_timeout = 15.0
 
-        if loop is None:
-            loop = asyncio.get_event_loop()
+        loop = get_running_loop(loop)
 
         self._closed = False
         if loop.get_debug():
             self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
-        self._conns = {}
+        self._conns = {}  # type: Dict[ConnectionKey, List[Tuple[ResponseHandler, float]]]  # noqa
         self._limit = limit
         self._limit_per_host = limit_per_host
-        self._acquired = set()
-        self._acquired_per_host = defaultdict(set)
-        self._keepalive_timeout = keepalive_timeout
+        self._acquired = set()  # type: Set[ResponseHandler]
+        self._acquired_per_host = defaultdict(set)  # type: DefaultDict[ConnectionKey, Set[ResponseHandler]]  # noqa
+        self._keepalive_timeout = cast(float, keepalive_timeout)
         self._force_close = force_close
 
         # {host_key: FIFO list of waiters}
-        self._waiters = defaultdict(deque)
+        self._waiters = defaultdict(deque)  # type: ignore
 
         self._loop = loop
         self._factory = functools.partial(ResponseHandler, loop=loop)
@@ -199,10 +231,10 @@ class BaseConnector:
         # start cleanup closed transports task
         self._cleanup_closed_handle = None
         self._cleanup_closed_disabled = not enable_cleanup_closed
-        self._cleanup_closed_transports = []
+        self._cleanup_closed_transports = []  # type: List[Optional[asyncio.Transport]]  # noqa
         self._cleanup_closed()
 
-    def __del__(self, _warnings=warnings):
+    def __del__(self, _warnings: Any=warnings) -> None:
         if self._closed:
             return
         if not self._conns:
@@ -210,7 +242,7 @@ class BaseConnector:
 
         conns = [repr(c) for c in self._conns.values()]
 
-        self.close()
+        self._close()
 
         if PY_36:
             kwargs = {'source': self}
@@ -226,19 +258,32 @@ class BaseConnector:
             context['source_traceback'] = self._source_traceback
         self._loop.call_exception_handler(context)
 
-    def __enter__(self):
+    def __enter__(self) -> 'BaseConnector':
+        warnings.warn('"witn Connector():" is deprecated, '
+                      'use "async with Connector():" instead',
+                      DeprecationWarning)
         return self
 
-    def __exit__(self, *exc):
+    def __exit__(self, *exc: Any) -> None:
         self.close()
 
+    async def __aenter__(self) -> 'BaseConnector':
+        return self
+
+    async def __aexit__(self,
+                        exc_type: Optional[Type[BaseException]]=None,
+                        exc_value: Optional[BaseException]=None,
+                        exc_traceback: Optional[TracebackType]=None
+                        ) -> None:
+        await self.close()
+
     @property
-    def force_close(self):
+    def force_close(self) -> bool:
         """Ultimately close connection on releasing if True."""
         return self._force_close
 
     @property
-    def limit(self):
+    def limit(self) -> int:
         """The total number for simultaneous connections.
 
         If limit is 0 the connector has no limit.
@@ -247,7 +292,7 @@ class BaseConnector:
         return self._limit
 
     @property
-    def limit_per_host(self):
+    def limit_per_host(self) -> int:
         """The limit_per_host for simultaneous connections
         to the same endpoint.
 
@@ -257,7 +302,7 @@ class BaseConnector:
         """
         return self._limit_per_host
 
-    def _cleanup(self):
+    def _cleanup(self) -> None:
         """Cleanup unused transports."""
         if self._cleanup_handle:
             self._cleanup_handle.cancel()
@@ -273,7 +318,8 @@ class BaseConnector:
                 for proto, use_time in conns:
                     if proto.is_connected():
                         if use_time - deadline < 0:
-                            transport = proto.close()
+                            transport = proto.transport
+                            proto.close()
                             if (key.is_ssl and
                                     not self._cleanup_closed_disabled):
                                 self._cleanup_closed_transports.append(
@@ -290,7 +336,8 @@ class BaseConnector:
             self._cleanup_handle = helpers.weakref_handle(
                 self, '_cleanup', timeout, self._loop)
 
-    def _drop_acquired_per_host(self, key, val):
+    def _drop_acquired_per_host(self, key: 'ConnectionKey',
+                                val: ResponseHandler) -> None:
         acquired_per_host = self._acquired_per_host
         if key not in acquired_per_host:
             return
@@ -299,7 +346,7 @@ class BaseConnector:
         if not conns:
             del self._acquired_per_host[key]
 
-    def _cleanup_closed(self):
+    def _cleanup_closed(self) -> None:
         """Double confirmation for transport close.
         Some broken ssl servers may leave socket open without proper close.
         """
@@ -317,8 +364,12 @@ class BaseConnector:
                 self, '_cleanup_closed',
                 self._cleanup_closed_period, self._loop)
 
-    def close(self):
+    def close(self) -> Awaitable[None]:
         """Close all opened transports."""
+        self._close()
+        return _DeprecationWaiter(noop2())
+
+    def _close(self) -> None:
         if self._closed:
             return
 
@@ -326,7 +377,7 @@ class BaseConnector:
 
         try:
             if self._loop.is_closed():
-                return noop()
+                return
 
             # cancel cleanup task
             if self._cleanup_handle:
@@ -356,14 +407,14 @@ class BaseConnector:
             self._cleanup_closed_handle = None
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         """Is connector closed.
 
         A readonly property.
         """
         return self._closed
 
-    def _available_connections(self, key):
+    def _available_connections(self, key: 'ConnectionKey') -> int:
         """
         Return number of available connections taking into account
         the limit, limit_per_host and the connection key.
@@ -379,19 +430,23 @@ class BaseConnector:
             # check limit per host
             if (self._limit_per_host and available > 0 and
                     key in self._acquired_per_host):
-                available = self._limit_per_host - len(
-                    self._acquired_per_host.get(key))
+                acquired = self._acquired_per_host.get(key)
+                assert acquired is not None
+                available = self._limit_per_host - len(acquired)
 
         elif self._limit_per_host and key in self._acquired_per_host:
             # check limit per host
-            available = self._limit_per_host - len(
-                self._acquired_per_host.get(key))
+            acquired = self._acquired_per_host.get(key)
+            assert acquired is not None
+            available = self._limit_per_host - len(acquired)
         else:
             available = 1
 
         return available
 
-    async def connect(self, req, traces, timeout):
+    async def connect(self, req: 'ClientRequest',
+                      traces: List['Trace'],
+                      timeout: 'ClientTimeout') -> Connection:
         """Get from pool or create new connection."""
         key = req.connection_key
         available = self._available_connections(key)
@@ -433,7 +488,7 @@ class BaseConnector:
 
         proto = self._get(key)
         if proto is None:
-            placeholder = _TransportPlaceholder()
+            placeholder = cast(ResponseHandler, _TransportPlaceholder())
             self._acquired.add(placeholder)
             self._acquired_per_host[key].add(placeholder)
 
@@ -469,7 +524,7 @@ class BaseConnector:
         self._acquired_per_host[key].add(proto)
         return Connection(self, key, proto, self._loop)
 
-    def _get(self, key):
+    def _get(self, key: 'ConnectionKey') -> Optional[ResponseHandler]:
         try:
             conns = self._conns[key]
         except KeyError:
@@ -480,7 +535,8 @@ class BaseConnector:
             proto, t0 = conns.pop()
             if proto.is_connected():
                 if t1 - t0 > self._keepalive_timeout:
-                    transport = proto.close()
+                    transport = proto.transport
+                    proto.close()
                     # only for SSL transports
                     if key.is_ssl and not self._cleanup_closed_disabled:
                         self._cleanup_closed_transports.append(transport)
@@ -494,7 +550,7 @@ class BaseConnector:
         del self._conns[key]
         return None
 
-    def _release_waiter(self):
+    def _release_waiter(self) -> None:
         """
         Iterates over all waiters till found one that is not finsihed and
         belongs to a host that has available connections.
@@ -518,7 +574,8 @@ class BaseConnector:
                     waiter.set_result(None)
                     return
 
-    def _release_acquired(self, key, proto):
+    def _release_acquired(self, key: 'ConnectionKey',
+                          proto: ResponseHandler) -> None:
         if self._closed:
             # acquired connection is already released on connector closing
             return
@@ -533,7 +590,8 @@ class BaseConnector:
         else:
             self._release_waiter()
 
-    def _release(self, key, protocol, *, should_close=False):
+    def _release(self, key: 'ConnectionKey', protocol: ResponseHandler,
+                 *, should_close: bool=False) -> None:
         if self._closed:
             # acquired connection is already released on connector closing
             return
@@ -544,7 +602,8 @@ class BaseConnector:
             should_close = True
 
         if should_close or protocol.should_close:
-            transport = protocol.close()
+            transport = protocol.transport
+            protocol.close()
 
             if key.is_ssl and not self._cleanup_closed_disabled:
                 self._cleanup_closed_transports.append(transport)
@@ -558,48 +617,50 @@ class BaseConnector:
                 self._cleanup_handle = helpers.weakref_handle(
                     self, '_cleanup', self._keepalive_timeout, self._loop)
 
-    async def _create_connection(self, req, traces, timeout):
+    async def _create_connection(self, req: 'ClientRequest',
+                                 traces: List['Trace'],
+                                 timeout: 'ClientTimeout') -> ResponseHandler:
         raise NotImplementedError()
 
 
 class _DNSCacheTable:
 
-    def __init__(self, ttl=None):
-        self._addrs_rr = {}
-        self._timestamps = {}
+    def __init__(self, ttl: Optional[float]=None) -> None:
+        self._addrs_rr = {}  # type: Dict[Tuple[str, int], Tuple[Iterator[Dict[str, Any]], int]]  # noqa
+        self._timestamps = {}  # type: Dict[Tuple[str, int], float]
         self._ttl = ttl
 
-    def __contains__(self, host):
+    def __contains__(self, host: object) -> bool:
         return host in self._addrs_rr
 
-    def add(self, host, addrs):
-        self._addrs_rr[host] = (cycle(addrs), len(addrs))
+    def add(self, key: Tuple[str, int], addrs: List[Dict[str, Any]]) -> None:
+        self._addrs_rr[key] = (cycle(addrs), len(addrs))
 
         if self._ttl:
-            self._timestamps[host] = monotonic()
+            self._timestamps[key] = monotonic()
 
-    def remove(self, host):
-        self._addrs_rr.pop(host, None)
+    def remove(self, key: Tuple[str, int]) -> None:
+        self._addrs_rr.pop(key, None)
 
         if self._ttl:
-            self._timestamps.pop(host, None)
+            self._timestamps.pop(key, None)
 
-    def clear(self):
+    def clear(self) -> None:
         self._addrs_rr.clear()
         self._timestamps.clear()
 
-    def next_addrs(self, host):
-        loop, length = self._addrs_rr[host]
+    def next_addrs(self, key: Tuple[str, int]) -> List[Dict[str, Any]]:
+        loop, length = self._addrs_rr[key]
         addrs = list(islice(loop, length))
         # Consume one more element to shift internal state of `cycle`
         next(loop)
         return addrs
 
-    def expired(self, host):
+    def expired(self, key: Tuple[str, int]) -> bool:
         if self._ttl is None:
             return False
 
-        return self._timestamps[host] + self._ttl < monotonic()
+        return self._timestamps[key] + self._ttl < monotonic()
 
 
 class TCPConnector(BaseConnector):
@@ -627,12 +688,19 @@ class TCPConnector(BaseConnector):
     loop - Optional event loop.
     """
 
-    def __init__(self, *, verify_ssl=True, fingerprint=None,
-                 use_dns_cache=True, ttl_dns_cache=10,
-                 family=0, ssl_context=None, ssl=None, local_addr=None,
-                 resolver=None, keepalive_timeout=sentinel,
-                 force_close=False, limit=100, limit_per_host=0,
-                 enable_cleanup_closed=False, loop=None):
+    def __init__(self, *, verify_ssl: bool=True,
+                 fingerprint: Optional[bytes]=None,
+                 use_dns_cache: bool=True, ttl_dns_cache: int=10,
+                 family: int=0,
+                 ssl_context: Optional[SSLContext]=None,
+                 ssl: Union[None, bool, Fingerprint, SSLContext]=None,
+                 local_addr: Optional[str]=None,
+                 resolver: Optional[AbstractResolver]=None,
+                 keepalive_timeout: Union[None, float, object]=sentinel,
+                 force_close: bool=False,
+                 limit: int=100, limit_per_host: int=0,
+                 enable_cleanup_closed: bool=False,
+                 loop: Optional[asyncio.AbstractEventLoop]=None):
         super().__init__(keepalive_timeout=keepalive_timeout,
                          force_close=force_close,
                          limit=limit, limit_per_host=limit_per_host,
@@ -647,28 +715,30 @@ class TCPConnector(BaseConnector):
 
         self._use_dns_cache = use_dns_cache
         self._cached_hosts = _DNSCacheTable(ttl=ttl_dns_cache)
-        self._throttle_dns_events = {}
+        self._throttle_dns_events = {}  # type: Dict[Tuple[str, int], EventResultOrError]  # noqa
         self._family = family
         self._local_addr = local_addr
 
-    def close(self):
+    def close(self) -> Awaitable[None]:
         """Close all ongoing DNS calls."""
         for ev in self._throttle_dns_events.values():
             ev.cancel()
 
-        super().close()
+        return super().close()
 
     @property
-    def family(self):
+    def family(self) -> int:
         """Socket family like AF_INET."""
         return self._family
 
     @property
-    def use_dns_cache(self):
+    def use_dns_cache(self) -> bool:
         """True if local DNS caching is enabled."""
         return self._use_dns_cache
 
-    def clear_dns_cache(self, host=None, port=None):
+    def clear_dns_cache(self,
+                        host: Optional[str]=None,
+                        port: Optional[int]=None) -> None:
         """Remove specified host/port or clear all dns local cache."""
         if host is not None and port is not None:
             self._cached_hosts.remove((host, port))
@@ -678,7 +748,10 @@ class TCPConnector(BaseConnector):
         else:
             self._cached_hosts.clear()
 
-    async def _resolve_host(self, host, port, traces=None):
+    async def _resolve_host(self,
+                            host: str, port: int,
+                            traces: Optional[List['Trace']]=None
+                            ) -> List[Dict[str, Any]]:
         if is_ip_address(host):
             return [{'hostname': host, 'host': host, 'port': port,
                      'family': self._family, 'proto': 0, 'flags': 0}]
@@ -744,7 +817,9 @@ class TCPConnector(BaseConnector):
 
         return self._cached_hosts.next_addrs(key)
 
-    async def _create_connection(self, req, traces, timeout):
+    async def _create_connection(self, req: 'ClientRequest',
+                                 traces: List['Trace'],
+                                 timeout: 'ClientTimeout') -> ResponseHandler:
         """Create connection.
 
         Has same keyword arguments as BaseEventLoop.create_connection.
@@ -760,7 +835,7 @@ class TCPConnector(BaseConnector):
 
     @staticmethod
     @functools.lru_cache(None)
-    def _make_ssl_context(verified):
+    def _make_ssl_context(verified: bool) -> SSLContext:
         if verified:
             return ssl.create_default_context()
         else:
@@ -771,7 +846,7 @@ class TCPConnector(BaseConnector):
             sslcontext.set_default_verify_paths()
             return sslcontext
 
-    def _get_ssl_context(self, req):
+    def _get_ssl_context(self, req: 'ClientRequest') -> Optional[SSLContext]:
         """Logic to get the correct SSL context
 
         0. if req.ssl is false, return None
@@ -804,7 +879,8 @@ class TCPConnector(BaseConnector):
         else:
             return None
 
-    def _get_fingerprint(self, req):
+    def _get_fingerprint(self,
+                         req: 'ClientRequest') -> Optional['Fingerprint']:
         ret = req.ssl
         if isinstance(ret, Fingerprint):
             return ret
@@ -813,14 +889,18 @@ class TCPConnector(BaseConnector):
             return ret
         return None
 
-    async def _wrap_create_connection(self, *args,
-                                      req, timeout,
-                                      client_error=ClientConnectorError,
-                                      **kwargs):
+    async def _wrap_create_connection(
+            self, *args: Any,
+            req: 'ClientRequest',
+            timeout: 'ClientTimeout',
+            client_error: Type[Exception]=ClientConnectorError,
+            **kwargs: Any) -> Tuple[asyncio.Transport, ResponseHandler]:
         try:
             with CeilTimeout(timeout.sock_connect):
-                return await self._loop.create_connection(*args, **kwargs)
-        except certificate_errors as exc:
+                return cast(
+                    Tuple[asyncio.Transport, ResponseHandler],
+                    await self._loop.create_connection(*args, **kwargs))
+        except cert_errors as exc:
             raise ClientConnectorCertificateError(
                 req.connection_key, exc) from exc
         except ssl_errors as exc:
@@ -828,8 +908,14 @@ class TCPConnector(BaseConnector):
         except OSError as exc:
             raise client_error(req.connection_key, exc) from exc
 
-    async def _create_direct_connection(self, req, traces, timeout,
-                                        *, client_error=ClientConnectorError):
+    async def _create_direct_connection(
+            self,
+            req: 'ClientRequest',
+            traces: List['Trace'],
+            timeout: 'ClientTimeout',
+            *,
+            client_error: Type[Exception]=ClientConnectorError
+    ) -> Tuple[asyncio.Transport, ResponseHandler]:
         sslcontext = self._get_ssl_context(req)
         fingerprint = self._get_fingerprint(req)
 
@@ -837,16 +923,20 @@ class TCPConnector(BaseConnector):
             # Cancelling this lookup should not cancel the underlying lookup
             #  or else the cancel event will get broadcast to all the waiters
             #  across all connections.
+            host = req.url.raw_host
+            assert host is not None
+            port = req.port
+            assert port is not None
             hosts = await asyncio.shield(self._resolve_host(
-                req.url.raw_host,
-                req.port,
+                host,
+                port,
                 traces=traces), loop=self._loop)
         except OSError as exc:
             # in case of proxy it is not ClientProxyConnectionError
             # it is problem of resolving proxy ip itself
             raise ClientConnectorError(req.connection_key, exc) from exc
 
-        last_exc = None
+        last_exc = None  # type: Optional[Exception]
 
         for hinfo in hosts:
             host = hinfo['host']
@@ -876,16 +966,24 @@ class TCPConnector(BaseConnector):
 
             return transp, proto
         else:
+            assert last_exc is not None
             raise last_exc
 
-    async def _create_proxy_connection(self, req, traces, timeout):
-        headers = {}
+    async def _create_proxy_connection(
+            self,
+            req: 'ClientRequest',
+            traces: List['Trace'],
+            timeout: 'ClientTimeout'
+    ) -> Tuple[asyncio.Transport, ResponseHandler]:
+        headers = {}  # type: Dict[str, str]
         if req.proxy_headers is not None:
-            headers = req.proxy_headers
+            headers = req.proxy_headers  # type: ignore
         headers[hdrs.HOST] = req.headers[hdrs.HOST]
 
+        url = req.proxy
+        assert url is not None
         proxy_req = ClientRequest(
-            hdrs.METH_GET, req.proxy,
+            hdrs.METH_GET, url,
             headers=headers,
             auth=req.proxy_auth,
             loop=self._loop,
@@ -927,7 +1025,9 @@ class TCPConnector(BaseConnector):
             conn = Connection(self, key, proto, self._loop)
             proxy_resp = await proxy_req.send(conn)
             try:
-                conn._protocol.set_response_params()
+                protocol = conn._protocol
+                assert protocol is not None
+                protocol.set_response_params()
                 resp = await proxy_resp.start(conn)
             except BaseException:
                 proxy_resp.close()
@@ -938,11 +1038,14 @@ class TCPConnector(BaseConnector):
                 conn._transport = None
                 try:
                     if resp.status != 200:
+                        message = resp.reason
+                        if message is None:
+                            message = RESPONSES[resp.status][0]
                         raise ClientHttpProxyError(
                             proxy_resp.request_info,
                             resp.history,
                             status=resp.status,
-                            message=resp.reason,
+                            message=message,
                             headers=resp.headers)
                     rawsock = transport.get_extra_info('socket', default=None)
                     if rawsock is None:
@@ -976,19 +1079,23 @@ class UnixConnector(BaseConnector):
     loop - Optional event loop.
     """
 
-    def __init__(self, path, force_close=False, keepalive_timeout=sentinel,
-                 limit=100, limit_per_host=0, loop=None):
+    def __init__(self, path: str, force_close: bool=False,
+                 keepalive_timeout: Union[object, float, None]=sentinel,
+                 limit: int=100, limit_per_host: int=0,
+                 loop: Optional[asyncio.AbstractEventLoop]=None) -> None:
         super().__init__(force_close=force_close,
                          keepalive_timeout=keepalive_timeout,
                          limit=limit, limit_per_host=limit_per_host, loop=loop)
         self._path = path
 
     @property
-    def path(self):
+    def path(self) -> str:
         """Path to unix socket."""
         return self._path
 
-    async def _create_connection(self, req, traces, timeout):
+    async def _create_connection(self, req: 'ClientRequest',
+                                 traces: List['Trace'],
+                                 timeout: 'ClientTimeout') -> ResponseHandler:
         try:
             with CeilTimeout(timeout.sock_connect):
                 _, proto = await self._loop.create_unix_connection(
@@ -996,4 +1103,4 @@ class UnixConnector(BaseConnector):
         except OSError as exc:
             raise ClientConnectorError(req.connection_key, exc) from exc
 
-        return proto
+        return cast(ResponseHandler, proto)
